package Data::Tree;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Tree - Class for working with Data::Trees

=head1 SYNOPSIS

 use Data::Tree;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;

#-- Custom application modules ------------------------------------------------#
use Data::Path::Parser;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $message = Data::Tree->new();
  Purpose   : Create new empty instance of the Mentat message
  Arguments : None
  Returns   : Reference to the new instance of the Mentat message
  Throws    : Croaks, if invoked on object

=cut

sub new()
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);

    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item unserialize($) [PUBLIC, STATIC]

  Usage     : my $message = Data::Tree->unserialize($string);
  Purpose   : Recover serialized instance of the Data::Tree from given string
  Arguments : STRING $string - string containing serialized instance
  Returns   : Reference to the recovered instance of the Mentat message
  Throws    : Croaks, if invoked on object

=cut

sub unserialize($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($string,) = @_;

    my $message = $class->new(decode_json($string));
    return $message;
}

=item from_file($) [PUBLIC, STATIC]

  Usage     : my $message = Data::Tree->from_file($filename);
  Purpose   : Recover serialized instance of the Mentat message from given file
  Arguments : STRING $filename - name of the file containing serialized instance
  Returns   : Reference to the recovered instance of the Mentat message
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses unserialize() method
  See Also  : unserialize() method

=cut

sub from_file($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($filename,) = @_;
    confess "Message file name must be given as argument" unless $filename;
    confess "Message file '$filename' is not ordinary file" unless -f $filename;

    # Load message from external file to string and untaint it
    open (my $mf, $filename) or die "Can`t open message file '$filename'\n";
    my $message = join("", <$mf>);
    close($mf);
    if ($message =~ /^(.+)$/s) { $message = $1; }

    # Unserialize the message from the string
    return $class->unserialize($message);
}

#-------------------------------------------------------------------------------

=item hash() [PUBLIC]

  Usage     : my $hash = $message->hash();
  Purpose   : Get the message structure hash
  Arguments : None
  Returns   : HASH REFERENCE $hash
  Throws    : Croaks, if invoked on class

=cut

sub hash()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return $self->{HASH};
}

#-------------------------------------------------------------------------------

=item iterate(;$$$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub iterate(;$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($pattern, $nodemode, $fullmode, $iterspec) = @_;

    unless ($self->{ITER}) {
        $self->{ITER} = [];
        $self->_prepare_iterator($self->{HASH},'', $fullmode, $iterspec);
        $self->{ITER} = [sort {$a->[0] cmp $b->[0]} @{$self->{ITER}}]
    }

    my $next;

    while ($next = shift(@{$self->{ITER}})) {
        my ($dpath, $node) = @$next;

        next if ($pattern) and not $dpath =~ /$pattern/;

        return ($dpath, $$node) unless $nodemode;
        return ($dpath, $node);
    }
    $self->{ITER} = undef;
    return (undef, undef);
}

#-------------------------------------------------------------------------------

=item path_nodes($) [PUBLIC]

  Usage     : my $nodes = $message->path_nodes($dpath);
  Purpose   : Get all nodes on given message path
  Arguments : STRING $dpath - message path to the nodes (MANDATORY)
  Returns   : reference to array of values
  Throws    : Croaks, if invoked on class
              Croaks, if not given message path
              Croaks, if node on given message path does not exist
  Comments  : Internally uses private methods _path_nodes()
  See Also  : Data::Tree::Path::Parser class

=cut

sub path_nodes($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath,) = @_;

    # Storage for nodes in current processing round
    my $nodes_a = [$self->{HASH}];

    # Storage for nodes in next processing round
    my $nodes_b = [];

    # Process the given MPath chunk by chunk
    my ($name, $type, $index, $tmp_val);
    do {
        ($name, $type, $index, $dpath) = Data::Path::Parser->parse_next_chunk($dpath);
        $name = ($type eq 'attr') ? '@' . $name : $name;

        # Check each node in the storage
        foreach my $node (@$nodes_a) {
            # Node must be a hash reference, otherwise it does not make any sense
            # asking for '$name', and of course '$name' key must exist
            next unless (ref $node eq 'HASH' and exists $node->{$name});

            # If the MPath contained index part
            if ($index) {
                # Node must be array reference, otherwise it does not make any sense
                # and of course array element at '$index' must exist
                next unless (ref $node->{$name} eq 'ARRAY' and $node->{$name}->[$index-1]);

                # We must pass everything by reference
                $tmp_val = (ref $node->{$name}->[$index-1]) ? $node->{$name}->[$index-1] : \$node->{$name}->[$index-1];

                # Let the node be handled in next round
                push(@$nodes_b, $tmp_val);
            }
            # We do not care about any particular index
            else {
                # If there are multiple subnodes under current node`s speciffic key, we must
                # add all of them
                if (ref $node->{$name} eq 'ARRAY') {
                    # C-like for loop must be used in this case, because we must pass everything by reference
                    for (my $i = 0; $i < scalar(@{$node->{$name}}); $i++) {
                        # We must pass everything by reference
                        $tmp_val = (ref $node->{$name}->[$i]) ? $node->{$name}->[$i] : \$node->{$name}->[$i];

                        # Let the node be handled in next round
                        push(@$nodes_b, $tmp_val);
                    }
                }
                else {
                    # We must pass everything by reference
                    $tmp_val = (ref $node->{$name}) ? $node->{$name} : \$node->{$name};

                    # Let the node be handled in next round
                    push(@$nodes_b, $tmp_val);
                }
            }
        }

        return undef unless scalar $nodes_b;
        $nodes_a = $nodes_b; $nodes_b = [];

    } while ($dpath);

    return $nodes_a;
}

=item path_node($) [PUBLIC]

  Usage     : my $node = $message->path_node($dpath);
  Purpose   : Get first node on given message path
  Arguments : STRING $dpath - message path to the nodes [MANDATORY]
  Returns   : STRING $value - node value (SUCCESS)
              undef (NODE HAS NO VALUE)
  Throws    : Croaks, if invoked on class
              Croaks, if not given message path
              Same as path_nodes() method
  Comments  : Internally uses path_nodes() method
  See Also  : path_nodes() method, Data::Tree::Path::Parser class

=cut

sub path_node($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath,) = @_;
    croak ((caller(0))[3] . ": invalid usage, message path must be given as argument") unless defined($dpath);

    # Get all matching nodes
    my $nodes = $self->path_nodes($dpath);

    # But return only the first one
    return shift @$nodes;
}

=item path_node_set($$) [PUBLIC]

  Usage     : $message->path_node_set($dpath, $value);
  Purpose   : If it does not exist, append new node to the Mentat message on the given path and set it`s value
  Arguments : STRING $dpath - path in the Mentat message structure (MANDATORY)
              STRING $value - value of the object (MANDATORY)
  Returns   : Nothing
  Throws    : Croaks, if invoked on class
              Croaks, if not given message path and value to set
              Croaks, if message path is syntactically incorrect
  Comments  : Internally uses private recursive method _append_node()
  See Also  : _append_node() method, Data::Tree::Path::Parser class

=cut

sub path_node_set($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath, $value) = @_;

    # Recursively parse the given MPath and create the inner HASH structure
    $self->_append_node($self->{HASH}, '', $dpath, $value);
}

=item path_node_delete($) [PUBLIC]

  Usage     : my $node = $message->path_delete($dpath);
  Purpose   : Deletes node on given message path
  Arguments : string $dpath - message path to the node (MANDATORY)
  Returns   : Nothing
  Throws    : Croaks, if invoked on class
              Croaks, if not given message path
              Croaks, if node on given message path does not exist
  Comments  : Internally uses private method _path_node()
  See Also  : _path_node() method, Data::Tree::Path::Parser class

=cut

sub path_node_delete($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath,) = @_;

    # Storage for nodes in current processing round
    my $nodes_a = [$self->{HASH}];

    # Storage for nodes in next processing round
    my $nodes_b = [];

    # Parse the given MPath in one call
    my @chunks = Data::Path::Parser->parse($dpath);

    my ($name, $type, $index, $tmp_val);
    for (my $i = 0; $i < scalar(@chunks); $i++) {
        ($name, $type, $index) = @{$chunks[$i]};
        $name = ($type eq 'attr') ? '@' . $name : $name;

        # Check each node in the storage
        foreach my $node (@$nodes_a) {

            # Node must be a hash reference, otherwise it does not make any sense
            # asking for '$name', and of course '$name' key must exist
            next unless (ref $node eq 'HASH' and exists $node->{$name});

            # If the MPath contained index part
            if ($index) {
                # '*' and '#' as index value are forbidden for deletion
                next if $index eq '*' or $index eq '#';

                # Node must be array reference, otherwise it does not make any sense
                # and of course array element at '$index' must exist
                next unless (ref $node->{$name} eq 'ARRAY' and $node->{$name}->[$index-1]);

                # End of the MPath, do the removal
                if (($i+1) == scalar(@chunks)) {
                    splice(@{$node->{$name}}, ($index-1), 1);
                }
                else {
                    # We must pass everything by reference
                    $tmp_val = (ref $node->{$name}->[$index-1]) ? $node->{$name}->[$index-1] : \$node->{$name}->[$index-1];

                    # Let the node be handled in next round
                    push(@$nodes_b, $tmp_val);
                }
            }
            # We do not care about any particular index
            else {
                # If there are multiple subnodes under current node`s speciffic key, we must
                # add all of them
                if (ref $node->{$name} eq 'ARRAY') {
                    # End of the MPath, do the removal
                    if (($i+1) == scalar(@chunks)) {
                        delete $node->{$name};
                    }
                    else {
                        # C-like for loop must be used in this case, because we must pass everything by reference
                        for (my $i = 0; $i < scalar(@{$node->{$name}}); $i++) {
                            # We must pass everything by reference
                            $tmp_val = (ref $node->{$name}->[$i]) ? $node->{$name}->[$i] : \$node->{$name}->[$i];

                            # Let the node be handled in next round
                            push(@$nodes_b, $tmp_val);
                        }
                    }
                }
                else {
                    # End of the MPath, do the removal
                    if (($i+1) == scalar(@chunks)) {
                        delete $node->{$name};
                    }
                    else {
                        # We must pass everything by reference
                        $tmp_val = (ref $node->{$name}) ? $node->{$name} : \$node->{$name};

                        # Let the node be handled in next round
                        push(@$nodes_b, $tmp_val);
                    }
                }
            }
        }

        last unless scalar $nodes_b;
        $nodes_a = $nodes_b; $nodes_b = [];
    }
}

=item path_node_exists($) [PRIVATE]

  Usage     : my $result = $message->path_node_exists($dpath);
  Purpose   : Tests, if the node on given message path exists
  Arguments : STRING $dpath - message path to the node [MANDATORY]
  Returns   : INTEGER 1 (SUCCESS)
              INTEGER 0 (FAILURE)
  Throws    : Croaks, if invoked on class
              Croaks, if not given any message path to test
  Comments  : Internally uses path_node() method
  See Also  : path_node() method, Data::Tree::Path::Parser class

=cut

sub path_node_exists($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath,) = @_;
    croak ((caller(0))[3] . ": invalid usage, message path must be given as argument") unless defined($dpath);

    # Try to find node located on the given message path
    my $node = $self->path_node($dpath);

    # If the node on given data path was not found, return undef
    return 0 unless $node;
    return 1;
}

#-------------------------------------------------------------------------------

=item path_values($) [PUBLIC]

  Usage     : my $value = $message->path_values($dpath);
  Purpose   : Gets values of all nodes on given message path
  Arguments : STRING $dpath - message path to the nodes (MANDATORY)
  Returns   : reference to array of values
  Throws    : Croaks, if invoked on class
              Croaks, if not given message path
              Croaks, if node on given message path does not exist
  Comments  : Internally uses private methods _path_nodes() and _get_child_text_node()
  See Also  : Data::Tree::Path::Parser class

=cut

sub path_values($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath,) = @_;

    my $nodes = $self->path_nodes($dpath);
    my $values = [];
    map { push(@$values, ((ref $_ eq 'SCALAR')?$$_:$_)); } @$nodes;
    return $values;
}

=item path_value($) [PUBLIC]

  Usage     : my $value = $message->path_value($dpath);
  Purpose   : Get value of first attribute or element node on given message path
  Arguments : STRING $dpath - message path to the nodes [MANDATORY]
  Returns   : STRING $value - node value (SUCCESS)
              undef (NODE HAS NO VALUE)
  Throws    : Croaks, if invoked on class
              Croaks, if not given message path
              Same as path_value() method
  Comments  : Internally uses path_values() method
  See Also  : path_values() method, Data::Tree::Path::Parser class

=cut

sub path_value($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath,) = @_;
    croak ((caller(0))[3] . ": invalid usage, message path must be given as argument") unless defined($dpath);

    # Get all mathing values
    my $values = $self->path_values($dpath);

    # But return only the first one
    return shift @$values;
}

=item path_value_set($$) [PUBLIC]

  Usage     : my $value = $message->path_value_set($dpath, $value);
  Purpose   : Set and return value of first node on given message path
  Arguments : STRING $dpath - message path to the node (MANDATORY)
              STRING $value - New node value (MANDATORY)
  Returns   : Nothing
  Throws    : Croaks, if invoked on class
              Croaks, if not given message path
              Croaks, if node on given message path does not exist
  Comments  : Internally uses private methods _path_node() and _get_child_text_node()
  See Also  : Data::Tree::Path::Parser class

=cut

sub path_value_set($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dpath, $value) = @_;

    my $node = $self->path_node($dpath);
    confess "Node on path '$dpath' does not exist" unless $node;

    if (ref $node eq 'SCALAR') {
        $$node = $value;
    }
    else {
        confess "Suspicious attempt to assign scalar value instead of previous HASH or ARRAY reference";
    }
}

#-------------------------------------------------------------------------------

=item serialize() [PUBLIC]

  Usage     : my $string = $message->serialize();
  Purpose   : Export Mentat message into string
  Arguments : None
  Returns   : STRING $string - string containing serialized Mentat message
  Throws    : Croaks, if invoked on class

=cut

sub serialize()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    #return encode_json($self->{HASH});
    return JSON->new->utf8->allow_blessed->convert_blessed->encode($self->{HASH})
}

=item to_string(;$) [PUBLIC]

  Usage     : my $string = $message->to_string();
  Purpose   : Export Mentat message into readable string
  Arguments : BOOL $pretty - format message nicely (OPTIONAL)
  Returns   : STRING $string - string containing serialized Mentat message
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub to_string(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $pretty = (shift @_) ? 1 : 0;

    return to_json($self->{HASH}, {utf8 => 1, pretty => $pretty, allow_blessed => 1, convert_blessed => 1});
}

=item to_file($) [PUBLIC]

  Usage     : $message->to_file($filename);
  Purpose   : Store Mentat message into external file
  Arguments : STRING $filename - name of the file to which to store the message [MANDATORY]
  Returns   : Nothing
  Throws    : Croaks, if invoked on class
              Croaks, if given file could not be opened for writing

=cut

sub to_file($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($file_name,) = @_;
    croak ((caller(0))[3] . ": name of the file to store the message must be given") unless $file_name;

    open(my $hnd, ">$file_name") or croak ((caller(0))[3] . ": cannot open file '$file_name' to store Mentat message for writing");
    print $hnd $self->serialize();
    close($hnd);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init(;$) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Data::Tree instance
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($hash,) = @_;

    # Build default message structure
    unless (defined $hash) {
        $hash = {};
    }
    else {
        confess "Message structure must be given via HASH reference" unless ref $hash eq 'HASH';
    }

    # Structure of hash of hashes
    $self->{HASH} = $hash;

    return $self;
}

# _append_node($$$$) [PRIVATE]
#
#  Usage     : $self->_append_node($self->{HASH}, $dpath, $rest, $value);
#  Purpose   : Recursively walks through the given MPath and creates all necessary nodes, sets the given value to the last one
#  Arguments : HASH REFERENCE $current - current node, we are working with (MANDATORY)
#              STRING $dpath - already processed MPath (MANDATORY)
#              STRING $rest - rest of the MPath to process (MANDATORY)
#              STRING $value - value to assign to the last node in the path (MANDATORY)
#  Returns   : Nothing
#  Throws    : Croaks, if invoked on class
#              Croaks, if MPath was invalid
#  Comments  : This method is for internal use only, it is called from path_node_set() method
#  See Also  : path_node_set() method

sub _append_node($$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($current, $dpath, $rest, $value) = @_;

    # Parse next chunk from the rest of the given MPath
    my ($name, $type, $index, $ref);
    ($name, $type, $index, $rest) = Data::Path::Parser->parse_next_chunk($rest);
    $name = ($type eq 'attr') ? '@' . $name : $name;

    # If we are not at the end of the path, build the structure recursively
    if ($rest) {

        if (not $index and exists($current->{$name}) and (ref $current->{$name} eq 'ARRAY')) {
            # EITHER complain
            #confess "Structure at '$name' was already defined as array multivalue: " . Dumper($current);

            # OR handle
            $index = 1;
        }

        # We must handle multivalues separately
        if ($index) {
            # Perhaps the node on this path already exists and it is not a multivalue
            if (exists($current->{$name}) and not (ref $current->{$name} eq 'ARRAY')) {
                # EITHER complain
                #confess "Structure at '$name' was already defined and not as array";

                # OR handle
                my $tmp = $current->{$name}; $current->{$name} = [$tmp];
            }
            # Push new element to the end of existing array
            if ($index eq '*') {
                my $x = {}; push(@{$current->{$name}}, $x);
                $self->_append_node($x, Data::Path::Parser->join($dpath,$name, $index), $rest, $value);
            }
            # Work with the last element of existing array
            elsif ($index eq '#') {
                my $l = 0; $l = $#{$current->{$name}} if ref $current->{$name} eq 'ARRAY';
                $current->{$name}->[$l] = {} unless exists $current->{$name}->[$l];
                $self->_append_node($current->{$name}->[$l], Data::Path::Parser->join($dpath, $name, $index), $rest, $value);
            }
            # Work with the element on specific position within array
            else {
                $current->{$name}->[$index-1] = {} unless exists $current->{$name}->[$index-1];
                $self->_append_node($current->{$name}->[$index-1], Data::Path::Parser->join($dpath, $name, $index), $rest, $value);
            }
        }
        # Single values are easy to deal with
        else {
            if (exists($current->{$name})) {
                if (ref $current->{$name} eq 'ARRAY') {
                    # EITHER complain
                    #confess "Structure at '$name' was already defined as array multivalue: " . Dumper($current);

                    # OR handle
                    my $tmp = $current->{$name}->[0];
                    return $self->_append_node($tmp, Data::Path::Parser->join($dpath, $name, $index), $rest, $value);
                }
                croak ((caller(0))[3] . ": structure at '$name' was already defined and not as HASH: " . Dumper($current)) unless (ref $current->{$name} eq 'HASH');
            }
            $current->{$name} = {} unless exists $current->{$name};
            $self->_append_node($current->{$name}, Data::Path::Parser->join($dpath, $name), $rest, $value);
        }
    }
    # End of the path, assign the value
    else {
        if (not $index and exists($current->{$name}) and (ref $current->{$name} eq 'ARRAY')) {
            # EITHER complain
            #croak ((caller(0))[3] . ": structure at '$name' was already defined as array multivalue: " . Dumper($current));

            # OR handle
            $index = 1;
        }

        # We must handle multivalues separately
        if ($index) {
            # Perhaps the node on this MPath already exists and it is not a multivalue
            if (exists($current->{$name}) and not (ref $current->{$name} eq 'ARRAY')) {
                # EITHER complain
                #croak ((caller(0))[3] . ": structure at '$name' was already defined and not as ARRAY multivalue: " . Dumper($current));

                # OR handle
                my $tmp = $current->{$name}; $current->{$name} = [$tmp];
            }
            # Append new element to existing array
            if ($index eq '*') {
                push(@{$current->{$name}}, $value);
            }
            # Replace the value of last element in array
            elsif ($index eq '#') {
                my $l = 0; $l = $#{$current->{$name}} if ref $current->{$name} eq 'ARRAY';
                $current->{$name}->[$l] = $value;
            }
            # Put the value to specific position
            else {
                $current->{$name}->[$index-1] = $value;
            }
        }
        # Single values are easy to deal with
        else {
            $current->{$name} = $value;
        }
    }
}

# _prepare_iterator($$;$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _prepare_iterator($$;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($hash, $path, $fullmode, $iterspec) = @_;
    
    my ($mpath, $ref, $i);
    foreach my $key (keys(%$hash)) {
        # Handle subhashes
        if(ref $hash->{$key} eq 'HASH') {
            $mpath = Data::Path::Parser->join($path, $key);
            $ref = \$hash->{$key};
            if ($fullmode) {
                push(@{$self->{ITER}}, [$mpath, $ref]);
            }
            # Mechanism for stopping the recursion for given nodes
            if ($iterspec and $mpath =~ /$iterspec/) {
                push(@{$self->{ITER}}, [$mpath, $ref]);
                next;
            }
            $self->_prepare_iterator($hash->{$key}, Data::Path::Parser->join($path, $key), $fullmode, $iterspec);
        }
        elsif (ref $hash->{$key} eq 'ARRAY') {

            $mpath = Data::Path::Parser->join($path, $key);
            $ref = \$hash->{$key};
            push(@{$self->{ITER}}, [$mpath, $ref]) if $fullmode;

            $i = 0;
            foreach my $item (@{$hash->{$key}}) {
                $mpath = Data::Path::Parser->join($path, $key, $i+1);
                if (ref $item eq 'HASH') {
                    $ref = \$hash->{$key}->[$i];
                    if ($fullmode) {
                        push(@{$self->{ITER}}, [$mpath, $ref]);
                    }
                    # Mechanism for stopping the recursion for given nodes
                    if ($iterspec and $mpath =~ /$iterspec/) {
                        push(@{$self->{ITER}}, [$mpath, $ref]);
                        next;
                    }
                    $self->_prepare_iterator($item, $mpath, $fullmode, $iterspec);
                }
                else {
                    $ref = \$hash->{$key}->[$i];
                    push(@{$self->{ITER}}, [$mpath, $ref]);
                }
                ++$i;
            }
        }
        else {
            $ref = \$hash->{$key};
            $mpath = Data::Path::Parser->join($path, $key);
            push(@{$self->{ITER}}, [$mpath, $ref]);
        }
    }
}

=pod

=back

=cut

1;
