package Mentat::Storage::Mongo::Aggregator;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Storage::Mongo::Aggregator - Short module description

=head1 SYNOPSIS

 use Mentat::Storage::Mongo::Aggregator;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Digest::MD5  qw(md5_hex);
use GD::Graph::pie; #libgd-graph-perl
use GD::Graph::bars;
use GD::Graph::area;

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use MongoDB::BSON::Binary;

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Value::Convertor;
use Data::Cache::Folder;
use Mentat::Storage::Mongo;
use Mentat::Core;
use Color::Palette;

# TODO UZ NEVIM jestli to je z cesnet tools nebo repozitare
use HTML::Entities;


#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
# Define meaningful default configuration values
use constant DFLT_DATA_CACHE_FOLDER => "/var/mentat/cache";
use constant DFLT_DRAW_AXISLABEL    => 0;
use constant DFLT_PALETTE_LIMIT     => 100;
use constant DFLT_DATA_LIMIT        => 10;

use constant SECS_YEAR => 31556926;

use constant AGGR_SOURCE => "AGGR_SOURCE";
use constant AGGR_TARGET => "AGGR_TARGET";
use constant AGGR_CLASS  => "AGGR_CLASS";
use constant AGGR_ANAL   => "AGGR_ANAL";

use constant GRAPH_PIE => "pie";
use constant GRAPH_BAR => "bar";

use constant FILTER          => "filter";
use constant FILTER_ACTION   => "action";
use constant FILTER_IPSRC    => "source";
use constant FILTER_IPDST    => "destination";
use constant FILTER_DETECTOR => "detector";
use constant FILTER_CATEGORY => "category";
use constant FILTER_PAGE     => "page";
use constant FILTER_ADVANCED => "advanced";
use constant FILTER_DATETO   => "dateto";
use constant FILTER_DATEFROM => "datefrom";
use constant FILTER_IPOR     => "ipor";

# Setup global configurations
my $DRAW_AXISLABEL = DFLT_DRAW_AXISLABEL();
my $PALETTE_LIMIT  = DFLT_PALETTE_LIMIT();
my $DATA_LIMIT     = DFLT_DATA_LIMIT();

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 1;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Storage::Mongo::Aggregator->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Storage::Mongo::Aggregator reference on success, croaks on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    ### Creating Mentat.Storage.Mongo.Aggregator instance...
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item get_item_sets($) [PUBLIC]

  Usage     : my $data = $instance->get_item_sets(1);
  Purpose   : Get alert data
  Arguments : $rewrite - force rewrite cache 0/1
  Returns   : ARRAY REFERENCE $data containing the alert data
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub get_item_sets($)
{
    my $self = shift;
    my ($rewrite, $stats_only) = @_;
    confess "Instance method not invoked on object instance" unless blessed($self);
    $rewrite    = (defined $rewrite    && int($rewrite)    == 1) ? 1 : 0;
    $stats_only = (defined $stats_only && int($stats_only) == 1) ? 1 : 0;
    
    my $r = {};
    my ($items, $message, $error);
    
    # Unless we are supposed to rewrite them, load values from cache
    unless ($rewrite) {
        foreach my $key ('itemset-stat-nodenames',
                         'itemset-stat-nodesws',
                         'itemset-stat-categories',
                         'itemset-stat-detectors') {
            Log::Loger->notice(__PACKAGE__, "Fetching item stat set '".$key."' from cache");
            $r->{$key} = $self->{CACHE}->fetch($key);
        }
        unless ($stats_only) {
            foreach my $key (Mentat::Core::KEY_CACHE_NODENAMES(),
                             Mentat::Core::KEY_CACHE_NODESWS(),
                             Mentat::Core::KEY_CACHE_CATEGORIES(),
                             Mentat::Core::KEY_CACHE_DETECTORS(),
                             Mentat::Core::KEY_CACHE_NODECATEG(),
                             Mentat::Core::KEY_CACHE_SWCATEG(),
                             Mentat::Core::KEY_CACHE_EVENTLIB()) {
                Log::Loger->notice(__PACKAGE__, "Fetching item data set '".$key."' from cache");
                $r->{$key} = $self->{CACHE}->fetch($key);
            }
        }
    }

    #---------------------------------------------------------------------------
    # New itemset loading from statistics
    #---------------------------------------------------------------------------

    # Fetch single stat item sets (distinct values of single attribute from statistics collection)
    my $tsl = time() - (SECS_YEAR()/2);
    foreach my $sset (['itemset-stat-nodenames',  'overall.list_detectors'],
                      ['itemset-stat-nodesws',    'overall.list_analyzers'],
                      ['itemset-stat-categories', 'overall.list_categories'],
                      ['itemset-stat-detectors',  'overall.list_detectorsws']) {
        # In case of cache miss we must calculate the data and store it to cache
        unless (defined $r->{$sset->[0]}) {
            # Fetch single item sets
            eval {
                Log::Loger->notice(__PACKAGE__, "Single item stat set '".$sset->[0]."' missing, fetching (".$sset->[1].") from database (with TS >= '".$tsl."')");
                $items = $self->{SSTORAGE}->distinct($sset->[1], undef, {'ts' => {'$gte' => $tsl}});
            };
            if ($@) {
                $message = "Error in getting item stat set for '".$sset->[0]."' from db: '$@'";
                Log::Loger->error(__PACKAGE__, $message);
                push (@{$r->{error}}, $message);
            }

            # Store to cache
            unless (defined $items && @$items) {
                $message = "No item stat set for '".$sset->[0]."' has been retrieved from db.";
                Log::Loger->debug(__PACKAGE__, $message) if $DEVEL;
                push (@{$r->{info}}, $message);
            }
            else {
                $error = $self->_cache_data($sset->[0], $items, $rewrite);
                if ($error) {
                    push (@{$r->{error}}, $error);
                }
                else {
                    $message = "Successfully updated item stat set '".$sset->[0]."'.";
                    push (@{$r->{info}}, $message);
                }
            }
            $r->{$sset->[0]} = $items;
        }
    }

    #---------------------------------------------------------------------------
    # Old itemset loading from alerts collection
    #---------------------------------------------------------------------------
    unless ($stats_only) {
    
        # Fetch single item sets (distinct values of single attribute)
        foreach my $set ([Mentat::Core::KEY_CACHE_NODENAMES(),  'Node.Name'],
                         [Mentat::Core::KEY_CACHE_NODESWS(),    'Node.SW'],
                         [Mentat::Core::KEY_CACHE_CATEGORIES(), 'Category']) {
            # In case of cache miss we must calculate the data and store it to cache
            unless (defined $r->{$set->[0]}) {
                # Fetch single item sets
                eval {
                    Log::Loger->notice(__PACKAGE__, "Single item data set '".$set->[0]."' missing, fetching (".$set->[1].") from database");
                    $items = $self->{STORAGE}->distinct($set->[1]);
                };
                if ($@) {
                    $message = "Error in getting item data set for '".$set->[0]."' from db: '$@'";
                    Log::Loger->error(__PACKAGE__, $message);
                    push (@{$r->{error}}, $message);
                }

                # Store to cache
                unless (defined $items && @$items) {
                    $message = "No item data set for '".$set->[0]."' has been retrieved from db.";
                    Log::Loger->debug(__PACKAGE__, $message) if $DEVEL;
                    push (@{$r->{info}}, $message);
                }
                else {
                    $error = $self->_cache_data($set->[0], $items, $rewrite);
                    if ($error) {
                        push (@{$r->{error}}, $error);
                    }
                    else {
                        $message = "Successfully updated item data set '".$set->[0]."'.";
                        push (@{$r->{info}}, $message);
                    }
                }
                $r->{$set->[0]} = $items;
            }
        }

        # Fetch dependent item sets (existing combinations of two single attributes)
        foreach my $depset ([Mentat::Core::KEY_CACHE_DETECTORS(), Mentat::Core::KEY_CACHE_NODENAMES(), Mentat::Core::KEY_CACHE_NODESWS(),    'Node.Name', 'Node.SW'],
                            [Mentat::Core::KEY_CACHE_NODECATEG(), Mentat::Core::KEY_CACHE_NODENAMES(), Mentat::Core::KEY_CACHE_CATEGORIES(), 'Node.Name', 'Category'],
                            [Mentat::Core::KEY_CACHE_SWCATEG(),   Mentat::Core::KEY_CACHE_NODESWS(),   Mentat::Core::KEY_CACHE_CATEGORIES(), 'Node.SW',   'Category']) {
            # In case of cache miss we must calculate the data and store it to cache
            unless (defined $r->{$depset->[0]}) {
                my $keys_a = $r->{$depset->[1]};
                my $keys_b = $r->{$depset->[2]};

                $items = [];

                # Fetch dependent item sets
                my $db = $self->{STORAGE}->{DATABASE};# TODO!!!
                my $collection = $db->get_collection('alerts');# TODO!!!
                foreach my $k_a (@$keys_a) {
                    foreach my $k_b (@$keys_b) {
                        Log::Loger->notice(__PACKAGE__, "Dependent item set '".$depset->[3].":".$depset->[4]."' missing, fetching (".$k_a.":".$k_b.") from database");
                        my $cursor = $collection->query({$depset->[3] => $k_a, $depset->[4] => $k_b})->limit(1);
                        if (my $next = $cursor->next) {
                            push (@$items, "$k_a/$k_b");
                        }
                    }
                }

                # Store to cache
                unless (defined $items && @$items) {
                    $message = "No item data set for '".$depset->[0]."' has been retrieved from db.";
                    Log::Loger->debug(__PACKAGE__, $message) if $DEVEL;
                    push (@{$r->{info}}, $message);
                }
                else {
                    $error = $self->_cache_data($depset->[0], $items, $rewrite);
                    if ($error) {
                        push (@{$r->{error}}, $error);
                    }
                    else {
                        $message = "Successfully updated dependent item data set '".$depset->[0]."'.";
                        push (@{$r->{info}}, $message);
                    }
                }
                $r->{$depset->[0]} = $items;
            }
        }

        #---------------------------------------------------------------------------
        
        # Build event library, if necessary
        unless (defined $r->{Mentat::Core::KEY_CACHE_EVENTLIB()}) {
            # Fetch data
            my $event_lib = {};
            $self->{STORAGE}->change_collection('events');
            my ($events, $ecnt) = $self->{STORAGE}->find();

            # Build library
            foreach my $e (@$events) {
                $event_lib->{by_id}->{$e->{_id}}++;
                $event_lib->{by_analyzer}->{$e->{analyzer}}->{$e->{_id}}++;
                #$event_lib->{by_category}->{$e->{category}}->{$e->{_id}}++;
                $event_lib->{by_class}->{$e->{class}}->{$e->{_id}}++;
                $event_lib->{by_anl_cls}->{$e->{analyzer}}->{$e->{class}}->{$e->{_id}}++;
            }

            # Perform data postprocessing and sorting
            $event_lib->{by_id} = [sort keys %{$event_lib->{by_id}}];
            foreach my $type ('by_analyzer', 'by_category', 'by_class') {
                next unless exists $event_lib->{$type};
                foreach my $key (keys %{$event_lib->{$type}}) {
                    $event_lib->{$type}->{$key} = [sort keys %{$event_lib->{$type}->{$key}}];
                }
            }
            foreach my $type ('by_anl_cls') {
                next unless exists $event_lib->{$type};
                foreach my $keya (keys %{$event_lib->{$type}}) {
                    foreach my $keyb (keys %{$event_lib->{$type}->{$keya}}) {
                        $event_lib->{$type}->{$keya}->{$keyb} = [sort keys %{$event_lib->{$type}->{$keya}->{$keyb}}];
                    }
                }
            }

            # Store to cache
            unless (defined $event_lib && scalar keys %$event_lib) {
                $message = "No event library data for '".Mentat::Core::KEY_CACHE_EVENTLIB()."' has been retrieved from db.";
                Log::Loger->debug(__PACKAGE__, $message) if $DEVEL;
                push (@{$r->{info}}, $message);
            }
            else {
                $error = $self->_cache_data(Mentat::Core::KEY_CACHE_EVENTLIB(), $event_lib, $rewrite);
                if ($error) {
                    push (@{$r->{error}}, $error);
                }
                else {
                    $message = "Successfully updated event library data '".Mentat::Core::KEY_CACHE_EVENTLIB()."'.";
                    push (@{$r->{info}}, $message);
                }
            }
            $r->{Mentat::Core::KEY_CACHE_EVENTLIB()} = $event_lib;
        }
    }
    
    return $r;
}

=item get_dashboard_data($$$) [PUBLIC]

  Usage     : my $data = $instance->get_dashboard_data($query, $from, $to);
  Purpose   : Get dashboard data for given query and time thresholds
  Arguments : STRING $query -
              $from -
              $to -
  Returns   : HASH REFERENCE $data containing the dashboard data
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub get_dashboard_data($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($query, $from, $to) = @_;
    #confess "Query must be given as first argument" unless $query;
    confess "Beginning of the time interval must be given as second argument" unless $from;
    confess "End of the time interval must be given as third argument" unless $to;
    my ($data, $error);

    my $epoch_from = Value::Convertor->datestr_to_unixs($from);
    my $epoch_to = Value::Convertor->datestr_to_unixs($to);

    if ($epoch_from > $epoch_to)
    {
        $error = "Incorrect time interval $epoch_from > $epoch_to.\n";
        Log::Loger->error(__PACKAGE__, $error);
        die "$error\n";
    }

    $self->{FROM} = $from;
    $self->{TO} = $to;

    # retrieved data must be converted (IP address)
    eval
    {
        my $ip = $query;
        $ip =~ s/ //g;
        my ($type, $min, $max) = Value::Convertor->ipv4_to_bin($ip);
        $self->{FILTER_IP} = $ip;
    };

    eval
    {
        my $ip = $query;
        $ip =~ s/ //g;
        my ($type, $min, $max) = Value::Convertor->ipv6_to_bin($ip);
        $self->{FILTER_IP} = $ip;
    };

    # Generate the data key
    my $cache_key = $self->generate_data_key($query, $epoch_from, $epoch_to);
    $self->{CACHE_KEY} = $cache_key;

    # Query the cache for calculated key
    $data = $self->{CACHE}->fetch($cache_key);

    # In case of cache miss we must calculate the data and store it to cache
    unless ($data) {
        $data = $self->_get_dashboard_data($query, $epoch_from, $epoch_to);
        $error = $self->{CACHE}->cache($cache_key, $data);

        # Report the error
        if ($error) {
            Log::Loger->error(__PACKAGE__, "Cache error: $error");
            die "$error\n";
        }
    }
    return $data;
}

=item generate_data_key($) [PUBLIC]

  Usage     : my $cache_key = $self->generate_data_key($);
  Purpose   : Generate unique data key
  Arguments : STRING $query
              $from
              $to
  Returns   : STRING $key
  Throws    : Dies, if not invoked on instance

=cut

sub generate_data_key($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($query, $from, $to) = @_;

    my ($key, $t) = ("", undef);
    $t = Value::Convertor->unixs_to_datetime($to,1);
    $key .= $t->strftime('%Y%m%d%H%M%S') . "-";
    $t = Value::Convertor->unixs_to_datetime($from,1);
    $key .= $t->strftime('%Y%m%d%H%M%S');

    if (defined $query && length ($query) > 0)
    {
        $query =~ s/ //g;
        my ($type, $min, $max) = ("", "", "");

        eval {
            ($type, $min, $max) = Value::Convertor->ipv4_to_bin($query);
        };
        eval {
            ($type, $min, $max) = Value::Convertor->ipv6_to_bin($query);
        };

        if ($type eq Value::Convertor::IPV4ADDR_STRING()) {
            my ($type, $ip) = Value::Convertor->ipv4_to_int($query);
            $key .= "_$ip";
        }

        elsif ($type eq Value::Convertor::IPV4RNG_STRING()) {
            my ($min, $max) = Value::Convertor->ipv4rngstr_to_ints($query);
            $key .= "_$min-$max";
        }

        elsif ($type eq Value::Convertor::IPV4CIDR_STRING()) {
            my ($min, $max) = Value::Convertor->ipv4cidr_to_ints($query);
            $key .= "_$min-$max";
        }

        elsif ($type eq Value::Convertor::IPV6ADDR_STRING()) {
            my ($type, $ip) = Value::Convertor->ipv6_to_bigint($query);
            $key .= "_$ip";
        }

        elsif ($type eq Value::Convertor::IPV6CIDR_STRING()) {
            my ($min, $max) = Value::Convertor->ipv6cidr_to_bigints($query);
            $key .= "_$min-$max";
        }

        else {
            $key .= md5_hex($query);
        }

    }

    return $key;
}


=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat.Storage.Mongo.Aggregator instance...

    delete $self->{STORAGE};
    delete $self->{CACHE};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Storage::Mongo::Aggregator instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Storage::Mongo::Aggregator $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat.Storage.Mongo.Aggregator instance...
    my %config = @_;

    my $palette = Color::Palette->new();
    my @p = $palette->palette($PALETTE_LIMIT);

    $self->{STORAGE}  = Mentat::Storage::Mongo->new(%config);
    $self->{SSTORAGE} = Mentat::Storage::Mongo->new(database => $config{sdatabase}, collection => $config{scollection});
    $self->{CACHE}    = Data::Cache::Folder->new($config{cache},$config{ttl},$config{clear});
    $self->{PALETTE}  = \@p;

    return $self;
}

sub _cache_data($$$)
{
    my $self = shift;
    my ($key, $data, $rewrite) = @_;

    my $error;
    eval {
        $error = $self->{CACHE}->cache($key, $data, $rewrite);
        die "$error\n" if $error;
    };
    if ($@) {
        my $message = "Cache error for '$key': '$@'";
        Log::Loger->error(__PACKAGE__, $message);
    }
    return $error;
}

# _get_dashboard_data($$$) [PROTECTED]
#
#  Usage     : my $data = $self->_get_dashboard_data($query, $from, $to);
#  Purpose   : Get dashboard data for given query and time thresholds
#              (actually perform the aggregation)
#  Arguments : STRING $query
#              $from
#              $to
#  Returns   : HASH REFERENCE $data
#  Throws    : Dies, if not invoked on instance

sub _get_dashboard_data($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($query, $epoch_from, $epoch_to) = @_;

    my $timegr = $self->_calculate_granularity($epoch_from, $epoch_to);
    my ($dashboard_data, $from, $to, $match, $p_source) = ([], $self->{FROM}, $self->{TO}, '', '');

    ($match, $p_source) = $self->_process_query($query, $epoch_from, $epoch_to, AGGR_SOURCE());
    my $source =
    [
        {
            '$match' => $match,
        },
        {
            '$project' => { source => $p_source }
        },
        {
         '$unwind' => '$source'
    },
        {
         '$unwind' => '$source'
    },
        {
            '$group' => { _id => '$source', sum => {'$sum' => 1} }
        },
        {
            '$sort' => { sum => -1 }
        },
    ];

    ($match, $p_source) = $self->_process_query($query, $epoch_from, $epoch_to, AGGR_TARGET());

    my $target =
    [
        {
            '$match' => $match,
        },
        {
            '$project' => { source => $p_source }
        },
        {
         '$unwind' => '$source'
    },
        {
         '$unwind' => '$source'
    },
        {
            '$group' => { _id => '$source', sum => {'$sum' => 1} }
        },
        {
            '$sort' => { sum => -1 }
        },
    ];


    my $class2_pipeline;
    my $class2_granul;
    my ($g, $i) = ($epoch_from, 0);
    while ($g < $epoch_to)
    {
        my $tg = Value::Convertor->unixs_to_datetime($g,1);
        $class2_granul->{$i} = $tg->strftime($timegr->{pattern});
        $g = $g + $timegr->{granul};
        $i++;
    }

    ($match, $p_source) = $self->_process_query($query, $epoch_from, $epoch_to, AGGR_CLASS());

    $class2_pipeline =
        [
            {
                '$match' => $match,
            },
            {
                '$project' => {
                    source => $p_source,
                    time => '$ts_u',
                    timegr => {
                        '$divide' => [
                            {'$subtract' => [
                                {'$subtract' => ['$ts_u', $epoch_from]},
                                {'$mod' => [
                                    {'$subtract' => ['$ts_u', $epoch_from]},
                                    $timegr->{granul}
                                ]}
                            ]},
                            $timegr->{granul}
                        ]
                    }
                }
            },
            { '$unwind' => '$source' },
            {
                '$group' => {
                    _id => {
                        timegr => '$timegr',
                        source => '$source'
                    },
                    sum => {'$sum' => 1}
                }
            },

        ];

    my $analyzer2_pipeline;
    my $analyzer2_granul;
    ($g, $i) = ($epoch_from, 0);
    while ($g < $epoch_to)
    {
        my $tg = Value::Convertor->unixs_to_datetime($g,1);
        $analyzer2_granul->{$i} = $tg->strftime($timegr->{pattern});
        $g = $g + $timegr->{granul};
        $i++;
    }

    ($match, $p_source) = $self->_process_query($query, $epoch_from, $epoch_to, AGGR_ANAL());

    $analyzer2_pipeline =
        [
            {
                '$match' => $match,
            },
            {
                '$project' => {
                    source => $p_source,
                    time => '$ts_u',
                    timegr => {
                        '$divide' => [
                            {'$subtract' => [
                                {'$subtract' => ['$ts_u', $epoch_from]},
                                {'$mod' => [
                                    {'$subtract' => ['$ts_u', $epoch_from]},
                                    $timegr->{granul}
                                ]}
                            ]},
                            $timegr->{granul}
                        ]
                    }
                }
            },
            { '$unwind' => '$source' },
            {
                '$group' => {
                    _id => {
                        timegr => '$timegr',
                        source => '$source'
                    },
                    sum => {'$sum' => 1}
                }
            },

        ];

    my $detector_pipeline;
    my $detector_granul;
    ($g, $i) = ($epoch_from, 0);
    while ($g < $epoch_to)
    {
        my $tg = Value::Convertor->unixs_to_datetime($g,1);
        $detector_granul->{$i} = $tg->strftime($timegr->{pattern});
        $g = $g + $timegr->{granul};
        $i++;
    }

    ($match, $p_source) = $self->_process_query($query, $epoch_from, $epoch_to, AGGR_ANAL());

    $detector_pipeline =
        [
            {
                '$match' => $match,
            },
            {
                '$project' => {
                    source => $p_source,
###                    name => '$Alert.Analyzer.@name',
                    name => '$Node.SW',
                    time => '$ts_u',
                    timegr => {
                        '$divide' => [
                            {'$subtract' => [
                                {'$subtract' => ['$ts_u', $epoch_from]},
                                {'$mod' => [
                                    {'$subtract' => ['$ts_u', $epoch_from]},
                                    $timegr->{granul}
                                ]}
                            ]},
                            $timegr->{granul}
                        ]
                    }
                }
            },
            { '$unwind' => '$source' },
            { '$unwind' => '$name' },
            { '$unwind' => '$name' },
            {
                '$group' => {
                    _id => {
                        timegr => '$timegr',
                        source => {
                            '$concat' => ['$source', '/', '$name']
                        }
                    },
                    sum => {'$sum' => 1}
                }
            },

        ];

    my $data_def = [
       {
           table => {
                title_column => 'Source IP',
                title_page => 'Source',
                id => 'db-source',
           },
           image =>
           [
               {
                    records => $self->{CACHE_KEY}."-source",
                    filename => $self->{CACHE_KEY}."-pie-source1",
                    rest => 0,
                    x_label => "$from - $to",
                    y_label => '',
                    title_image => 'Source IP (TOP 10)',
                    graph_type => GRAPH_PIE(),
               },
               {
                    records => $self->{CACHE_KEY}."-source",
                    filename => $self->{CACHE_KEY}."-pie-source2",
                    rest => 1,
                    x_label => "$from - $to",
                    y_label => '',
                    title_image => 'Source IP (ALL)',
                    graph_type => GRAPH_PIE(),
               },
           ],
           type => 'ip',
           pipeline => $target,
           FILTER() => FILTER_IPSRC(),
           from => $from,
           to => $to,
           limit => 10,
       },

       {
           table => {
                title_column => 'Destination IP',
                title_page => 'Destination',
                id => 'db-destination',
           },
           image =>
           [
               {
                    records => $self->{CACHE_KEY}."-target",
                    filename => $self->{CACHE_KEY}."-pie-target1",
                    rest => 0,
                    x_label => "$from - $to",
                    y_label => '',
                    title_image => 'Destination IP (TOP 10)',
                    graph_type => GRAPH_PIE(),
               },
               {
                    records => $self->{CACHE_KEY}."-target",
                    filename => $self->{CACHE_KEY}."-pie-target2",
                    rest => 1,
                    x_label => "$from - $to",
                    y_label => '',
                    title_image => 'Destination IP (ALL)',
                    graph_type => GRAPH_PIE(),
               },
           ],
           type => 'ip',
           pipeline => $source,
           FILTER() => FILTER_IPDST(),
           from => $from,
           to => $to,
           limit => 10,
       },

       {
           table => {
                title_column => 'Category',
                title_page => 'Category',
                id => 'db-classification',
           },
           image =>
           [
               {
                    records => $self->{CACHE_KEY}."-class",
                    filename => $self->{CACHE_KEY}."-bar-class",
                    x_label => "$from - $to (step ".$timegr->{granul_title}.")",
                    y_label => '# OF alerts',
                    title_image => 'Category',
                    graph_type => GRAPH_BAR(),
               },
               {
                   records => $self->{CACHE_KEY}."-class",
                   filename => $self->{CACHE_KEY}."-pie-class",
                   x_label => "$from - $to",
                   y_label => '',
                   title_image => 'Category',
                   graph_type => GRAPH_PIE(),
               }
           ],
           type => '',
           pipeline => $class2_pipeline,
           granul => $class2_granul,
           FILTER() => FILTER_CATEGORY(),
           from => $from,
           to => $to,
           limit => 0,
       },

       {
           table => {
                title_column => 'Analyzer',
                title_page => 'Analyzer',
                id => 'db-analyzer',
           },
           image =>
           [
               {
                    records => $self->{CACHE_KEY}."-analyzer",
                    filename => $self->{CACHE_KEY}."-bar-analyzer",
                    x_label => "$from - $to (step ".$timegr->{granul_title}.")",
                    y_label => '# OF alerts',
                    title_image => 'Analyzer',
                    graph_type => GRAPH_BAR(),
               },
               {
                   records => $self->{CACHE_KEY}."-analyzer",
                   filename => $self->{CACHE_KEY}."-pie-analyzer",
                   x_label => "$from - $to",
                   y_label => '',
                   title_image => 'Analyzer',
                   graph_type => GRAPH_PIE(),
               }
           ],
           type => '',
           pipeline => $analyzer2_pipeline,
           granul => $analyzer2_granul,
           FILTER() => FILTER_DETECTOR(),
           from => $from,
           to => $to,
           limit => 0,
       },

       {
           table => {
                title_column => 'Detector',
                title_page => 'Detector',
                id => 'db-detector',
           },
           image =>
           [
               {
                    records => $self->{CACHE_KEY}."-detector",
                    filename => $self->{CACHE_KEY}."-bar-detector",
                    x_label => "$from - $to (step ".$timegr->{granul_title}.")",
                    y_label => '# OF alerts',
                    title_image => 'Detector',
                    graph_type => GRAPH_BAR(),
               },
               {
                   records => $self->{CACHE_KEY}."-detector",
                   filename => $self->{CACHE_KEY}."-pie-detector",
                   x_label => "$from - $to",
                   y_label => '',
                   title_image => 'Detector',
                   graph_type => GRAPH_PIE(),
               }
           ],
           type => '',
           pipeline => $detector_pipeline,
           granul => $detector_granul,
           FILTER() => FILTER_DETECTOR(),
           from => $from,
           to => $to,
           limit => 0,
       },

    ];

    foreach my $df (@$data_def)
    {
        my $dd;
        $dd->{table} = $df->{table};
        $dd->{image} = [];

        foreach my $image (@{$df->{image}})
        {
            my $key = $image->{records}."-".$image->{graph_type};

            my $records = $self->{CACHE}->fetch($key);
            unless ($records)
            {
                $records = $self->_get_data($df->{pipeline});
            }

            my $data = $self->_get_data_graph($df, $image, $records);

            my $dd_image;
            $dd_image = $image;
            $dd_image->{data} = $data->{graph};
            $dd_image->{palette} = $data->{palette};

            push (@{$dd->{image}}, $dd_image);
            $dd->{table}->{data} = $data->{table};
        }

        push (@$dashboard_data, $dd);
    }

    unless (defined $dashboard_data)
    {
        $dashboard_data = [];
    }

    return $dashboard_data;
}

# _calculate_granularity($$) [PROTECTED]
#
#  Usage     : my $granularity = $self->_calculate_granularity($epoch_from, $epoch_to);
#  Purpose   : Calculate data granularity for given aggregation interval
#  Arguments : INTEGER $epoch_from - start of the aggregation interval [MANDATORY]
#              INTEGER $epoch_to - end of the aggregation interval [MANDATORY]
#  Returns   : HASH REFERENCE $granularity
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _calculate_granularity($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($epoch_from, $epoch_to) = @_;
    confess "Start of the interval must be given as first argument" unless $epoch_from;
    confess "End of the interval must be given as second argument" unless $epoch_to;

    my $interval = $epoch_to - $epoch_from;
    my $granul = 0;
    my $granul_title = "";

    if (($interval > 1) && ($interval <= 30))
    {
        $granul = 1;
        $granul_title = "1 second";
    }
    elsif (($interval > 30) && ($interval <= 60))
    {
        $granul = 2;
        $granul_title = "2 seconds";
    }
    elsif (($interval > 60) && ($interval <= 120))
    {
        $granul = 5;
        $granul_title = "5 seconds";
    }
    elsif (($interval > 120) && ($interval <= 240))
    {
        $granul = 10;
        $granul_title = "10 seconds";
    }
    elsif (($interval > 240) && ($interval <= 480))
    {
        $granul = 30;
        $granul_title = "30 seconds";
    }
    elsif (($interval > 480) && ($interval <= 960))
    {
        $granul = 60;
        $granul_title = "1 minute";
    }
    elsif (($interval > 960) && ($interval <= 1800))
    {
        $granul = 2*60;
        $granul_title = "2 minutes";
    }
    elsif (($interval > 1800) && ($interval <= 3600))
    {
        $granul = 5*60;
        $granul_title = "5 minutes";
    }
    elsif (($interval > 3600) && ($interval <= 2*3600))
    {
        $granul = 10*60;
        $granul_title = "10 minutes";
    }
    elsif (($interval > 2*3600) && ($interval <= 24*3600))
    {
        $granul = 3600;
        $granul_title = "1 hour";
    }
    elsif (($interval > 24*3600) && ($interval <= 48*3600))
    {
        $granul = 2*3600;
        $granul_title = "2 hours";
    }
    elsif (($interval > 48*3600) && ($interval <= 96*3600))
    {
        $granul = 4*3600;
        $granul_title = "4 hours";
    }
    elsif (($interval > 4*86400) && ($interval <= 7*86400))
    {
        $granul = 12*3600;
        $granul_title = "12 hours";
    }
    elsif (($interval > 7*86400) && ($interval <= 21*86400))
    {
        $granul = 1*86400;
        $granul_title = "1 day";
    }
    elsif (($interval > 21*86400) && ($interval <= 30*86400))
    {
        $granul = 2*86400;
        $granul_title = "2 days";
    }
    elsif (($interval > 30*86400) && ($interval <= 60*86400))
    {
        $granul = 4*86400;
        $granul_title = "4 days";
    }
    elsif (($interval > 60*86400) && ($interval <= 120*86400))
    {
        $granul = 7*86400;
        $granul_title = "7 days";
    }
    elsif (($interval > 120*86400) && ($interval <= 180*86400))
    {
        $granul = 14*86400;
        $granul_title = "14 days";
    }
    else
    {
        $granul = 30*86400;
        $granul_title = "30 days";
    }


    my $first = Value::Convertor->unixs_to_datetime($epoch_from,1);
    my $second = Value::Convertor->unixs_to_datetime($epoch_from+$granul,1);
    my $last = Value::Convertor->unixs_to_datetime($epoch_to,1);
    my $pattern = ''; # pattern
    my $p = '';       # temp pattern

    # RRRR-MM-DD
    if ($first->strftime('%F') ne $last->strftime('%F'))
    {

        if ($first->strftime('%d') ne $last->strftime('%d'))
        {
            $p = '%d.';
        }

        if ($first->strftime('%m') ne $last->strftime('%m'))
        {
            $p = '%d.%m.';
        }

        if ($first->strftime('%Y') ne $last->strftime('%Y'))
        {
            $p = '%d.%m.%Y';
        }
    }

    # HH:MM:SS
    if ($first->strftime('%T') eq '00:00:00' && $last->strftime('%T') eq '00:00:00' && $second->strftime('%T') eq '00:00:00')
    {
        $pattern = $p;
    }
    # SS
    elsif ($first->strftime('%S') eq '00' && $last->strftime('%S') eq '00' && $second->strftime('%S') eq '00')
    {
        $pattern = (length ($p) > 0)?$p.' %R':$p;
    }
    else
    {
        $pattern = (length ($p) > 0)?$p.' %T':'%T';
    }

    return { granul => $granul, granul_title => $granul_title, pattern => $pattern };
}

# _process_query($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _process_query($)
{

    my $self    = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

    my ($query, $epoch_from, $epoch_to, $aggr) = @_;

    my $match_ts = { 'ts_u' => { '$gte' => $epoch_from,  '$lte' => $epoch_to}};
    my ($match, $source, $ipv6) = ($match_ts, '', 0);

    my ($type, $min, $max);
    if (defined $query && length ($query) > 0) {

        $query =~ s/ //g;

        eval {
            ($type, $min, $max) = Value::Convertor->ipv4_to_bin($query);
        };

        eval {
            ($type, $min, $max) = Value::Convertor->ipv6_to_bin($query);
            $ipv6 = 1;
        };

    }

    $source = '$Target.IP4.ip' if ($ipv6 == 0 && $aggr eq AGGR_SOURCE());
    $source = '$Target.IP6.ip' if ($ipv6 == 1 && $aggr eq AGGR_SOURCE());
    $source = '$Source.IP4.ip' if ($ipv6 == 0 && $aggr eq AGGR_TARGET());
    $source = '$Source.IP6.ip' if ($ipv6 == 1 && $aggr eq AGGR_TARGET());
    ### $source = '$Alert.Analyzer.@analyzerid' if ($aggr eq AGGR_ANAL());
    $source = '$Node.Name' if ($aggr eq AGGR_ANAL());

    ### $source = '$Alert.Classification.@text' if ($aggr eq AGGR_CLASS());
    $source = '$Category' if ($aggr eq AGGR_CLASS());

    if (defined $type) {

        # --- IPV4 ---

        if ($type eq Value::Convertor::IPV4ADDR_STRING())
        {
            $min = MongoDB::BSON::Binary->new(data => $min, subtype => 0);

            if ($aggr eq AGGR_SOURCE())
            {
                $match = {'Source.IP4.ip' => $min, $match_ts};
            }
            elsif ($aggr eq AGGR_TARGET())
            {
                $match = {'Target.IP4.ip' => $min, $match_ts};
            }
            elsif (($aggr eq AGGR_ANAL()) || ($aggr eq AGGR_CLASS()))
            {
                $match = {
                    '$and' => [
                        { '$or' => [
                            {'Source.IP4.ip' => $min},
                            {'Target.IP4.ip' => $min}
                        ]},
                        $match_ts
                    ]};
            }
        }

        if ($type eq Value::Convertor::IPV4CIDR_STRING() || $type eq Value::Convertor::IPV4RNG_STRING())
        {
            $min = MongoDB::BSON::Binary->new(data => $min, subtype => 0);
            $max = MongoDB::BSON::Binary->new(data => $max, subtype => 0);

            if ($aggr eq AGGR_SOURCE())
            {
                $match = {
                    '$and' => [
                        {'Source.IP4.min' => { '$gte' => $min}},
                        {'Source.IP4.max' => { '$lte' => $max}},
                        $match_ts
                    ]
                };
            }

            elsif ($aggr eq AGGR_TARGET())
            {
                $match = {
                    '$and' => [
                        {'Target.IP4.min' => { '$gte' => $min}},
                        {'Target.IP4.max' => { '$lte' => $max}},
                        $match_ts
                    ]
                };
            }

            elsif (($aggr eq AGGR_ANAL()) || ($aggr eq AGGR_CLASS()))
            {
                $match =
                {
                    '$or' => [
                        {
                            '$and' => [
                                {'Source.IP4.min' => { '$gte' => $min}},
                                {'Source.IP4.max' => { '$lte' => $max}},
                                $match_ts
                            ]
                        },
                        {
                            '$and' => [
                                {'Target.IP4.min' => { '$gte' => $min}},
                                {'Target.IP4.max' => { '$lte' => $max}},
                                $match_ts
                            ]
                        },
                       ]
                };
            }

        }

        # --- IPV6 ---

        if ($type eq Value::Convertor::IPV6ADDR_STRING())
        {
            $min = MongoDB::BSON::Binary->new(data => $min, subtype => 0);

            if ($aggr eq AGGR_SOURCE())
            {
                $match = {'Source.IP6.ip' => $min, $match_ts};
            }
            elsif ($aggr eq AGGR_TARGET())
            {
                $match = {'Target.IP6.ip' => $min, $match_ts};
            }
            elsif (($aggr eq AGGR_ANAL()) || ($aggr eq AGGR_CLASS()))
            {
                $match = {
                    '$and' => [
                        { '$or' => [
                            {'Source.IP6.ip' => $min},
                            {'Target.IP6.ip' => $min}
                        ]},
                        $match_ts
                    ]};
            }
        }

        if ($type eq Value::Convertor::IPV6CIDR_STRING() || $type eq Value::Convertor::IPV6RNG_STRING())
        {
            $min = MongoDB::BSON::Binary->new(data => $min, subtype => 0);
            $max = MongoDB::BSON::Binary->new(data => $max, subtype => 0);

            if ($aggr eq AGGR_SOURCE())
            {
                $match = {
                    '$and' => [
                        {'Source.IP6.min' => { '$gte' => $min}},
                        {'Source.IP6.max' => { '$lte' => $max}},
                        $match_ts
                    ]
                };
            }

            elsif ($aggr eq AGGR_TARGET())
            {
                $match = {
                    '$and' => [
                        {'Target.IP6.min' => { '$gte' => $min}},
                        {'Target.IP6.max' => { '$lte' => $max}},
                        $match_ts
                    ]
                };
            }

            elsif (($aggr eq AGGR_ANAL()) || ($aggr eq AGGR_CLASS()))
            {
                $match =
                {
                    '$or' => [
                        {
                            '$and' => [
                                {'Source.IP6.min' => { '$gte' => $min}},
                                {'Source.IP6.max' => { '$lte' => $max}},
                                $match_ts
                            ]
                        },
                        {
                            '$and' => [
                                {'Target.IP6.min' => { '$gte' => $min}},
                                {'Target.IP6.max' => { '$lte' => $max}},
                                $match_ts
                            ]
                        },
                       ]
                };
            }

        }
    }

    return ($match, $source);
}

# _get_data($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _get_data($)
{
    my $self    = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

    my ($pipeline) = @_;
    my $records;

    if (defined $pipeline)
    {
        eval {
            $records = $self->{STORAGE}->aggregate($pipeline);
        };
        if ($@)
        {
            my $message = "Unable to load data from mongo: $@";
            Log::Loger->error(__PACKAGE__, $message);
            die;
        }
    }
    else
    {
        Log::Loger->debug(__PACKAGE__, "Pipeline does not exists") if ($DEVEL);
    }

    return $records;
}

# _get_data_graph($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _get_data_graph($$$)
{
    my $self    = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

    my ($df, $image, $records) = @_;
    my $graph_type = $image->{graph_type};
    my $result = undef;

    if ($graph_type eq GRAPH_BAR())
    {
        my $graph = $self->_get_data_graph_bar($df, $records);

        my $key = $image->{records}."-".GRAPH_PIE();
        my $object = $graph->{GRAPH_PIE()}->{records};
        $self->{CACHE}->cache($key, $object) if (defined $object);

        $result = $self->_data_bar($df, $graph) if (defined $image->{records});
    }

    elsif ($graph_type eq GRAPH_PIE())
    {
        my $graph = $self->_get_data_graph_pie($df, $records);

        my $key = $image->{records}."-".GRAPH_PIE();
        my $object = $graph->{GRAPH_PIE()}->{records};
        $self->{CACHE}->cache($key, $object) if (defined $object);

        $result = $self->_data_pie($df, $graph, $image->{rest}) if (defined $image->{records});
    }

    return $result;

}

# _get_data_graph_bar($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _get_data_graph_bar($$)
{
    my $self    = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

    my ($df, $records) = @_;
    my $granul = $df->{granul};

    my $graph;
    $graph->{GRAPH_BAR()}->{totalsum} = 0;

    while ( my ($g, $date) = each(%$granul) )
    {
        my $g_key = $g.";".$date;
        $graph->{GRAPH_BAR()}->{granul}->{$g_key}->{""} = 0;
    }

    my $SUCCESS = 0;
    foreach my $record (@$records)
    {
        my $id = $record->{_id}->{source};
        my $g_date = $granul->{$record->{_id}->{timegr}};
        my $g_num = $record->{_id}->{timegr};
        my $g_key = $g_num.";".$g_date;
        my $sum = $record->{sum};

        $graph->{GRAPH_BAR()}->{granul}->{$g_key}->{$id} = $sum;
        $graph->{GRAPH_BAR()}->{records}->{$id} = 0 if (! defined $graph->{GRAPH_BAR()}->{records}->{$id});
        $graph->{GRAPH_BAR()}->{records}->{$id} += $sum;
        $graph->{GRAPH_BAR()}->{totalsum} += $sum;

        $SUCCESS = 1;
    }

    # data retrieved from db
    if ($SUCCESS == 1)
    {
        # -------------------------
        my $pie = $graph->{GRAPH_BAR()}->{records};
        #  Purpose   : Help sort a pie by the pie 'value', not the 'key'. Values are returned in ascending numeric order (lowest to highest).
        my $pieValueAscendingNum = sub { $pie->{$a} <=> $pie->{$b}; };
        #  Purpose   : Help sort a pie by the pie 'value', not the 'key'. Values are returned in descending numeric order (highest to lowest).
        my $pieValueDescendingNum = sub { $pie->{$b} <=> $pie->{$a}; };

        foreach my $id (sort $pieValueDescendingNum (keys(%$pie)))
        {
            push (@{$graph->{GRAPH_PIE()}->{records}}, {_id => $id, sum => $pie->{$id}});
        }
        # -------------------------

        return $graph;
    }
    else
    {
        Log::Loger->debug(__PACKAGE__, "Any data retrieved from db") if ($DEVEL);
        return undef;
    }
}

# _get_data_graph_pie($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _get_data_graph_pie($$)
{
    my $self    = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

    my ($df, $records) = @_;

    my $limit = (defined $df->{limit} && $df->{limit} >= 0)?$df->{limit}:$DATA_LIMIT;
    my $count = (defined $records)?@$records:0;

    my $graph;
    $graph->{GRAPH_PIE()}->{totalsum} = 0;
    $graph->{GRAPH_PIE()}->{head} = 0;
    $graph->{GRAPH_PIE()}->{tail} = 0;
    $graph->{GRAPH_PIE()}->{count} = $count;
    $graph->{GRAPH_PIE()}->{records} = $records;

    my $i = 0;
    my $SUCCESS = 0;
    foreach my $record (@$records)
    {
        my $sum = $record->{sum};
        $graph->{GRAPH_PIE()}->{head} += $sum if ($i < $limit);
        $graph->{GRAPH_PIE()}->{tail} += $sum if ($i >= $limit);
        $graph->{GRAPH_PIE()}->{totalsum} += $sum;
        $i++;
        $SUCCESS = 1;
    }

    # data retrieved from db
    if ($SUCCESS == 1)
    {
        return $graph;
    }
    else
    {
        Log::Loger->debug(__PACKAGE__, "Any data retrieved from db") if ($DEVEL);
        return undef;
    }
}

# _data_bar($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _data_bar($$)
{
    my $self    = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

    my ($df, $graph) = @_;

    my (@data, @table, @legend, @palette);

    my $days        = $graph->{GRAPH_BAR()}->{granul};
    my $records     = $graph->{GRAPH_BAR()}->{records};
    my $totalsum    = $graph->{GRAPH_BAR()}->{totalsum};

    my $limit = (defined $df->{limit} && $df->{limit} >= 0)?$df->{limit}:$DATA_LIMIT;

    #  Purpose   : Help sort a hash by the hash 'value', not the 'key'. Values are returned in ascending numeric order (lowest to highest).
    my $hashValueAscendingNum = sub { $records->{$a} <=> $records->{$b}; };
    #  Purpose   : Help sort a hash by the hash 'value', not the 'key'. Values are returned in descending numeric order (highest to lowest).
    my $hashValueDescendingNum = sub { $records->{$b} <=> $records->{$a}; };

    my @dd = sort keys %$days;
    my @d = sort keys %$days;

    @d = sort {
        my ($digits_from_a) = $a =~ /(\d+);/;
        my ($digits_from_b) = $b =~ /(\d+);/;
        $digits_from_a <=> $digits_from_b
    } @d;

    @dd = sort {
        my ($digits_from_a) = $a =~ /(\d+);/;
        my ($digits_from_b) = $b =~ /(\d+);/;
        $digits_from_a <=> $digits_from_b
    } @dd;

    s/\d+;// for @dd;
    push (@data, \@dd);

    my $i = 0;
    foreach my $id (sort $hashValueDescendingNum (keys(%$records)))
    {
        my $count = $records->{$id};
        my $color = "#".$self->{PALETTE}->[$i][1];
        my $filter_param = $df->{FILTER()};
        my $filter = {
            FILTER_ADVANCED() => 0,
            FILTER_ACTION()   => 'Search',
            FILTER_DATETO()   => $df->{to},
            FILTER_DATEFROM() => $df->{from},
            $filter_param => $id,
        };

        if ( defined $self->{FILTER_IP} && length ($self->{FILTER_IP}) > 0)
        {
            if ( $filter_param ne FILTER_IPSRC() && $filter_param ne FILTER_IPDST() )
            {
                $filter->{FILTER_IPSRC()}   = $self->{FILTER_IP};
                $filter->{FILTER_IPOR()}    = "on";
                $filter->{FILTER_IPDST()}   = $self->{FILTER_IP};
            }
            else
            {
                $filter->{($filter_param eq FILTER_IPSRC())?FILTER_IPDST():FILTER_IPSRC()} = $self->{FILTER_IP};
            }
        }

        my @array;
        for my $day (@d)
        {
            my $count = $days->{$day}->{$id};
            $count = 0 if (!defined $count);
            push (@array, $count);
        }

        push (@table, {color => $color , value => $id, count => $count, top => sprintf("%.2f%%", abs ((100/$totalsum)*$count)), all => "-", FILTER() => $filter});
        push (@palette, $color);
        push (@legend, ($DRAW_AXISLABEL == 1)?$id:"");
        push (@data, \@array);
        $i++;
    }

    if ($i <= $limit && $limit > 0)
    {
        push (@table, {color => "" , value => "The rest", count => 0, top => "-", all => "0.00%", FILTER() => undef});
    }

    return {graph => [\@legend, \@data], table => \@table, palette => \@palette};

}

# _data_pie($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _data_pie($$$)
{
    my $self    = shift;
    croak ((caller(0))[3] . ": instance method invoked on class") unless ref $self;

    my ($df, $graph, $rest_image) = @_;

    my $limit = (defined $df->{limit} && $df->{limit} >= 0)?$df->{limit}:$DATA_LIMIT;

    my @palette;

    my (@graph1, @graph2, @table);
    my @graph_array = (\@graph1, \@graph2);

    my $data = $graph->{GRAPH_PIE()};
    my $totalsum = (defined $data->{totalsum} && $data->{totalsum} > 0)?$data->{totalsum}:0;
    my $tail = (defined $data->{tail} && $data->{tail} > 0)?$data->{tail}:0;
    my $head = (defined $data->{head} && $data->{head} > 0)?$data->{head}:$data->{tail};
    my $count = (defined $data->{count} && $data->{count} > 0)?$data->{count}:0;
    my $records = $data->{records};
    my $i=0;


    foreach my $record (@$records)
    {

        my $value = $record->{_id};
    my $sum = $record->{sum};
        my $color = "#".$self->{PALETTE}->[$i][1];

        if ($i >= $limit && $limit != 0)
        {

            my $all = sprintf("%.2f%%", abs (((100/$totalsum)*$tail)));
            push (@table, {color => $color, value => "The rest", count => $tail, top => "-", all => $all, FILTER() => undef});

            if ($rest_image == 1)
            {
                push (@graph1, ($DRAW_AXISLABEL == 1)?"rest":"");
                push (@graph2, $tail);
                push (@palette, $color);
            }
            last;
        }
        else
        {
            eval
            {
                no warnings;
                $value = Value::Convertor->bin_to_ipv4str($value) if (($df->{type} eq 'ip') and (length ($value) > 0));
            };
            if ($@) {
                my $message = "Value: $value\nUnable convert query to ipv4: $@";
                Log::Loger->error(__PACKAGE__, $message);
            }


            my $filter_param = $df->{FILTER()};
            my $filter = {
                FILTER_ADVANCED() => 0,
                FILTER_ACTION()   => 'Search',
                FILTER_DATETO()   => $df->{to},
                FILTER_DATEFROM() => $df->{from},
                $filter_param => $value,
            };

            if ( defined $self->{FILTER_IP} && length ($self->{FILTER_IP}) > 0)
            {
                if ( $filter_param ne FILTER_IPSRC() && $filter_param ne FILTER_IPDST() )
                {
                    $filter->{FILTER_IPSRC()}   = $self->{FILTER_IP};
                    $filter->{FILTER_IPOR()}    = "on";
                    $filter->{FILTER_IPDST()}   = $self->{FILTER_IP};
                }
                else
                {
                    $filter->{($filter_param eq FILTER_IPSRC())?FILTER_IPDST():FILTER_IPSRC()} = $self->{FILTER_IP};
                }
            }

            my $sum = $record->{sum};
            my $top = sprintf("%.2f%%", abs ((100/$head)*$sum));
            my $all = sprintf("%.2f%%", abs (((100/$totalsum)*$sum)));
            push (@table, {color => $color , value => $value, count => $sum, top => $top, all => $all, FILTER() => $filter});
            push (@graph1, ($DRAW_AXISLABEL == 1)?$value:"");
            push (@graph2, $sum);
            push (@palette, $color);
        }

        $i++;
    }

    if ($count <= $limit && $limit != 0 && $count > 0)
    {
        my $color = "#".$self->{PALETTE}->[$i][1];
        if ($rest_image == 1)
        {
            push (@graph1, ($DRAW_AXISLABEL == 1)?"rest":"");
            push (@graph2, $tail);
            push (@palette, $color);
        }

        my $all = sprintf("%.2f%%", abs (((100/$totalsum)*$tail)));
        push (@table, {color => $color, value => "The rest", count => $tail, top => "-", all => $all, FILTER() => undef});

    }

    return {graph => \@graph_array, table => \@table, palette => \@palette};

}

=pod

=back

=cut

1;
