package Mentat::System;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::System - Short module description

=head1 SYNOPSIS

 use Mentat::System;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Cwd 'abs_path';
use File::stat;
use JSON;

#use Smart::Comments; #-+-> DEVEL ONLY <-+-#
use Data::Dumper; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

# Modules from Mentat package
use Mentat::Core qw(:func);

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION    = '0.1';
    $DEVEL      = 0;
    $ENV{PATH}  = '/bin:/usr/bin';
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item runlogs_load($) [PUBLIC,STATIC]

  Usage     : my $runlogs = Mentat::System->runlogs_load($runlog_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub runlogs_load($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Runlog directory must be given as argument\n" unless $dir;
    die "Runlog directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {};
    my @runlogs = $class->dir_load($dir, '.*\.rl$');

    my $data;
    foreach my $rl (@runlogs) {
        # $rl->[0] = file; $rl->[1] = path;
        $data = json_load($rl->[1]);
        $result->{$data->{app_name}} = $data;
    }

    return $result;
}

=item runlogs_clean($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->runlogs_clean($runlog_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub runlogs_clean($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Runlog directory must be given as argument\n" unless $dir;
    die "Runlog directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {};
    my @runlogs = $class->dir_load($dir, '.*\.rl$');

    foreach my $rl (@runlogs) {
        # $rl->[0] = file; $rl->[1] = path;
        $result->{$rl->[0]} = (unlink($rl->[1])) ? 1 : 0;
    }

    return $result;
}

#-------------------------------------------------------------------------------

=item logs_list($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->logs_list($log_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub logs_list($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Log directory must be given as argument\n" unless $dir;
    die "Log directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {};
    my @logs = $class->dir_load($dir, '.*\.log');

    my $st;
    foreach my $l (@logs) {
        $st = stat($l->[1]) or die "Unable to stat file '".$l->[1]."': '$!'";
        $result->{$l->[0]} = {
                'size'  => $st->size,
                'mode'  => $st->mode,
                'atime' => $st->atime,
                'mtime' => $st->mtime,
                'uid'   => $st->uid,
                'gid'   => $st->gid,
            };
    }

    return $result;
}

=item logs_clean($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->logs_clean($log_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub logs_clean($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Log directory must be given as argument\n" unless $dir;
    die "Log directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {};
    my @logs = $class->dir_load($dir, '.*\.log');

    my $r;
    foreach my $l (@logs) {
        # Truncate
        if ($l->[0] =~ /\.log$/) {
            if(open(F, ">", $l->[1])) {
                close(F);
                $result->{$l->[0]} = 1;
            }
            else {
                $result->{$l->[0]} = 0;
            }
        }
        # Unlink
        else {
            $r = unlink($l->[1]);
            if ($r) {
                $result->{$l->[0]} = 2;
            }
            else {
                $result->{$l->[0]} = 0;
            }
        }
    }

    return $result;
}

#-------------------------------------------------------------------------------

=item spool_list($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->spool_list($spool_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub spool_list($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Log directory must be given as argument\n" unless $dir;
    die "Log directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {};
    my @spools = $class->dir_load($dir, '^_|\.py$');

    my (@wdirs, $cnt);
    # _pickup, _storage, ...
    foreach my $s (@spools) {
        if (-d $s->[1]) {
            @wdirs = $class->dir_load($s->[1]);
            # incoming, pending, errors, ...
            foreach my $wd (@wdirs) {
                $cnt = $class->dir_file_count($wd->[1]);
                $result->{$s->[0]}->{$wd->[0]} = $cnt;
            }
        }
    }

    return $result;
}

=item spool_clean($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->spool_clean($spool_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub spool_clean($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Log directory must be given as argument\n" unless $dir;
    die "Log directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {};
    my @spools = $class->dir_load($dir, '^_|\.py$');

    my (@wdirs, @messages, $cnt, $r);
    # _pickup, _storage, ...
    foreach my $s (@spools) {
        if (-d $s->[1]) {
            @wdirs = $class->dir_load($s->[1]);

            $result->{$s->[0]}->{removed} = 0;
            $result->{$s->[0]}->{errors}  = 0;

            # incoming, pending, errors, ...
            foreach my $wd (@wdirs) {
                next unless $wd->[0] eq 'errors' or $wd->[0] eq 'pending';

                my @messages = $class->dir_load($wd->[1], '.*\.(idea|idmef)');
                foreach my $mf (@messages) {
                    $r = unlink($mf->[1]);
                    if ($r) { $result->{$s->[0]}->{removed}++; }
                    else    { $result->{$s->[0]}->{errors}++; }
                }
            }
        }
    }

    return $result;
}

#-------------------------------------------------------------------------------

=item script_list($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->script_list($script_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub script_list($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Script directory must be given as argument\n" unless $dir;
    die "Script directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result   = {};
    my @cronjobs = $class->dir_load('/etc/cron.d');
    my @scripts  = $class->dir_load($dir, '^cron_');

    # Build a list of currently active cron jobs
    my %crons = (); my $link;
    foreach my $cj (@cronjobs) {
        next unless -l $cj->[1];
        $link = readlink($cj->[1]);
        $crons{$link} = 1;
    }

    my $sname;
    foreach my $s (@scripts) {
        next if $s->[0] =~ /dpkg-old|dpkg-dist$/;
        if ($s->[0] =~ /^cron_(.*)/) {
            $sname = $1;
            $result->{$1} = {
                'cron'    => $s->[0],
                'enabled' => (($crons{$s->[1]}) ? 1 : 0),
            }
        }
    }

    return $result;
}

#-------------------------------------------------------------------------------

=item pidfile_list($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->pidfile_list($run_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub pidfile_list($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Run directory must be given as argument\n" unless $dir;
    die "Run directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {};
    my @pidfiles = $class->dir_load($dir, '.pid$');

    my ($pid, $running, $running_0);
    foreach my $pf (@pidfiles) {
        $pid = get_pid_from_file($pf->[1]);
        if ($pid) {
            $running   = is_running_by_pid($pid);
            $running_0 = (kill(0, $pid)) ? 1 : 0;
            $result->{$pf->[0]} = {
                'pid'       => $pid,
                'file'      => $pf->[1],
                'running'   => $running,
                'running_0' => $running_0,
            };
        }
        else {
            $result->{$pf->[0]} = {
                'error' => 'Malformed PID file',
            };
        }
    }

    return $result;
}

=item pidfile_clean($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->pidfile_clean($run_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub pidfile_clean($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Run directory must be given as argument\n" unless $dir;
    die "Run directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result = {};
    my @pidfiles = $class->dir_load($dir, '.pid$');

    my ($pid, $running, $r);
    foreach my $pf (@pidfiles) {
        $pid = get_pid_from_file($pf->[1]);
        if ($pid) {
            $running = is_running_by_pid($pid);

            if ($running) {
                push(@{$result->{kept}}, $pf->[0]);
            }
            else {
                $r = unlink($pf->[1]);
                if ($r) { push(@{$result->{removed}}, $pf->[0]); }
                else    { push(@{$result->{errors}}, $pf->[0]); }
            }
        }
        else {
            $result->{$pf->[0]} = {
                'error' => 'Malformed PID file',
            };
        }
    }

    return $result;
}

#-------------------------------------------------------------------------------

=item cache_list($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->cache_list($cache_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub cache_list($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;
    die "Cache directory must be given as argument\n" unless $dir;
    die "Cache directory '$dir' is not ordinary directory\n" unless -d $dir;

    my $result  = {
            'dir'   => $dir,
            'count' => 0,
            'size'  => 0,
        };
    my @cachefiles = $class->dir_load($dir);

    my $st;
    foreach my $cf (@cachefiles) {
        $st = stat($cf->[1]) or die "Unable to stat file '".$cf->[1]."': '$!'";
        $result->{count}++;
        $result->{size}     += $st->size;
        $result->{ts_newest} = $st->mtime if not $result->{ts_newest} or $result->{ts_newest} < $st->mtime;
        $result->{ts_oldest} = $st->mtime if not $result->{ts_oldest} or $result->{ts_oldest} > $st->mtime;
    }
    $result->{age_newest} = (time - $result->{ts_newest}) if $result->{ts_newest};
    $result->{age_oldest} = (time - $result->{ts_oldest}) if $result->{ts_oldest};

    return $result;
}

=item cache_clean($) [PUBLIC,STATIC]

  Usage     : my $result = Mentat::System->cache_clean($cache_dir);
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub cache_clean($;$)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir, $ttl) = @_;
    die "Cache directory must be given as argument\n" unless $dir;
    die "Cache directory '$dir' is not ordinary directory\n" unless -d $dir;

    $ttl = (60 * 60 * 24 * 7) unless defined $ttl;
    $ttl = int($ttl);

    my $result  = {
            'dir'     => $dir,
            'ttl'     => $ttl,
            'kept'    => 0,
            'removed' => 0,
            'errors'  => 0,
            'freed'   => 0,
        };
    my @cachefiles = $class->dir_load($dir);

    my ($st, $r);
    my $ct = time;
    foreach my $cf (@cachefiles) {
        $st = stat($cf->[1]) or die "Unable to stat file '".$cf->[1]."': '$!'";
        if (($st->mtime + $ttl) < $ct) {
            $r = unlink $cf->[1];
            if ($r) { $result->{removed}++; $result->{freed} += $st->size; }
            else    { $result->{errors}++; }
        }
        else {
            $result->{kept}++;
        }
    }

    return $result;
}

#sub rrds_list


#sub repair() (file and directory permissions, executability, ...)
#sub host_status() (disk free space, etc.)

#sub component_list()
#sub component_status() (check component status)
#sub component_stats() (dump and load component statistics)

#-------------------------------------------------------------------------------
# HELPERS
#-------------------------------------------------------------------------------

sub dir_load($;$)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir, $ptrn) = @_;

    opendir (DIR, $dir ) || die "Error in opening dir '$dir'\n";
    my ($file, $path);
    my @files = ();
    $ptrn = qr/$ptrn/ if $ptrn;
    while(($file = readdir(DIR))){
        next if $file =~ /^\.\.?$/;
        if ($ptrn) {
            next unless $file =~ /$ptrn/;
        }
        $path = join('/', $dir, $file);
        if ($file =~ /(.*)/) { $file = $1; }
        if ($path =~ /(.*)/) { $path = $1; }
        push(@files, [$file, $path]);
    }
    closedir(DIR);
    return @files;
}

sub dir_count($;$)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir, $ptrn) = @_;

    opendir (DIR, $dir ) || die "Error in opening dir '$dir'\n";
    my $file;
    my $cnt = 0;
    $ptrn = qr/$ptrn/ if $ptrn;
    while(($file = readdir(DIR))){
        next if $file =~ /^\.\.?$/;
        if ($ptrn) {
            next unless $file =~ /$ptrn/;
        }
        $cnt++;
    }
    closedir(DIR);
    return $cnt;
}

sub dir_file_count($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($dir,) = @_;

    my $cnt = `/usr/bin/find $dir -type f | wc -l`;
    return int($cnt);
}

# get_pid_from_file($)
#
#  Usage     : my $pid = get_pid_from_file($pid_file);
#  Purpose   : Load PID of the process from given file
#  Arguments :
#  Returns   :

sub get_pid_from_file($)
{
    my ($pid_file,) = @_;

    my $rv = open(PID_FILE, $pid_file) or return undef;
    my $line = <PID_FILE>;
    chomp($line);
    close(PID_FILE);

    return $1 if $line =~ /(\d+)/;
    return undef;
}

# is_running_by_pid($)
#
#  Usage     : my $rv = is_running_by_pid($pid);
#  Purpose   : Get number of running processes with given name
#  Arguments :
#  Returns   :
#  Throws    :

sub is_running_by_pid($)
{
    my ($pid,) = @_;

    #my $rv = `/bin/ps aux | /bin/grep "$pid" | /bin/grep -v grep | /usr/bin/wc -l`;
    my $rv = `/bin/ps aux | awk '{ print \$2 }' | /bin/grep "$pid" | /usr/bin/wc -l`;
    return int($rv);
}

=pod

=back

=cut

1;
